> module TypeCheck where
>
> -- import Debug.Trace
>
> import Data.List hiding (group)
> import Data.Maybe (fromJust)
> import Control.Monad.State

> import Options
> import Message hiding (concat, break)
> import AbstractSyntax

> type TC a = State TCState a
> type TypeError = Doc
> data TCState = TCState
>                  {  tccount   ::  !Int,
>                     tcsubst   ::  TSubst,
>                     tcerrors  ::  [TypeError]  }

> bind :: Int -> Type -> TC ()
> bind n t =
>   -- tError (text ("binding " ++ show n ++ " to ") <> ppType t) () >>
>   applytcsubst t >>= \t' -> modify (\s -> s { tcsubst = (n,t') : tcsubst s })

> fresh :: TC Int
> fresh = get >>= \s -> let n = tccount s + 1 in put (s { tccount = n }) >> return n

> applytcsubst :: Type -> TC Type
> applytcsubst t = gets tcsubst >>= \s -> return (subst [] s t)

> tError :: Loc -> TypeError -> a -> TC a
> tError l x te = modify (\s -> s { tcerrors = (pretty l <> text " " <> x) : tcerrors s }) >> return te

> loc = fst

About namespaces:

Value namespace:
Values are always lowercase.
Constructors are always uppercase.
Objects are values.

Type namespace:
Type variables are lowercase.
Type constructors are uppercase.
Classes are types.

Module namespace:
Modules are uppercase.

Signature namespace:
Signatures are uppercase.




Interestingly, we can reuse the language of module specifications
(used in signatures) for the type environment. Of course, the idea
is that like with expressions and values, only certain constructs
occur.

> type TypeEnv = [Spec]

I first thought that functors are not covered by this, but they
are. Functors are written like

  sig
    module F = functor (X : SX) -> sig fun f : X.T -> X.T end
  end

(for instance), i.e., they are just treated like modules.

A |TypeEnv| maps value variables to types, but also structure
variables to signatures.

The function |vLookup| looks up type variables and data constructors.

> vLookup :: LVar -> TypeEnv -> Maybe Type
> vLookup v0 te = msum (map (vLookup' v0) te)
>   where
>     vLookup' (_,v) (SpVal (_,w) t) | v == w  =  return t
>     vLookup' (_,v) (SpCon (_,c) t) | v == c  =  return t
>     vLookup' (_,v) (SpException ((_,c),Nothing))
>                                    | v == c  =  return tException
>     vLookup' (_,v) (SpException ((_,c),Just t))
>                                    | v == c  =  return (tFun [t] tException)
>     vLookup' _ _                             =  Nothing

The function |tLookup| looks up type constructors:

> tLookup :: LCon -> TypeEnv -> Maybe Type
> tLookup v0 te
>   | snd (splitTyName v0) /= ""         =  return (TyCon v0) -- we know what we're looking for
> tLookup v0 te = msum (map (tLookup' v0) te)
>   where
>     tLookup' (_,v) (SpType (_,c) t)
>       -- | v == fst (splitTyName c)       =  return t
>       | v == c                         =  return t
>     tLookup' (_,v) (SpData d@(_,c))
>       | v == fst (splitTyName d)       =  return (TyCon d)
>       | v == c                         =  return (TyCon d)
>     tLookup' _ _                       =  Nothing

About data types:

Data types cannot be identified by name only, because several
data types of the same name may exist in different scope, and
because of nested modules and functors we are in a less convenient
situation than for Haskell and cannot just assign entity names.
Our solution is to assign numbers to each data type, and
use these numbers to disambiguate. Splitting the number off a
type name is done by |splitTyName|.

The function |strLookup| looks up structure variables.

> strLookup :: LStrVar -> TypeEnv -> Maybe TypeEnv
> strLookup v0 te = msum (map (strLookup' v0) te)
>   where
>     strLookup' (_,v) (SpModule (_,w) (Sig s)) | v == w  =  return s
>     strLookup' _ _                                      =  Nothing

The function |tcLookup| looks up variables in a type substitution.

> tcLookup :: LTyVar -> TypeSubst -> Maybe Type
> tcLookup _ [] = Nothing
> tcLookup x@(_,v) (((_,w),t):r)
>   | v == w                   =  return t
>   | otherwise                =  tcLookup x r

> type TypeSubst
>   =  [(LTyVar, Type)]

> type TSubst
>   =  [(Int, Type)]

> builtin :: TyCon -> Type
> builtin x = TyCon (unknown,x) -- QualTyCon (StrVar "Builtin")

> tBool, tNat, tString, tException, tRef, tArray, tCont, tVoid, tValue, tUnit :: Type

> tBool       =  builtin iBool
> tNat        =  builtin iNat
> tString     =  builtin iString
> tException  =  builtin iException
> tRef        =  builtin iRef
> tArray      =  builtin iArray
> tCont       =  builtin iCont
> tVoid       =  builtin iVoid
> tValue      =  builtin iValue
> tUnit       =  tyRecord []

> iBool, iNat, iString, iException, iRef, iArray, iCont, iVoid, iValue :: String
> iBool       =  makeTyName 0 "Bool"
> iNat        =  makeTyName 0 "Nat"
> iString     =  makeTyName 0 "String"
> iException  =  makeTyName 0 "Exception"
> iRef        =  makeTyName 0 "Ref"
> iArray      =  makeTyName 0 "Array"
> iCont       =  makeTyName 0 "Cont"
> iVoid       =  makeTyName 0 "Void"
> iValue      =  makeTyName 0 "Value"

> tFun :: [Type] -> Type -> Type
> tFun ts t = TyFun (map (\s -> ForPat (PatTyped (PatVar (unknown,"_")) s)) ts) t


> typecheck           ::  TypeEnv -> Expr -> TC Type
> typecheck'          ::  TypeEnv -> Expr -> TC Type
> typecheckDecl       ::  TypeEnv -> Decl -> TC TypeEnv
> typecheckDecls      ::  TypeEnv -> [Decl] -> TC TypeEnv
> typecheckRow        ::  TypeEnv -> Row Expr -> TC TyRow
> typecheckArm        ::  Doc -> TypeEnv -> Type -> Type -> (Pat, Expr) -> TC Type
> typecheckAlt        ::  TypeEnv -> Alt -> Type -> TC Spec
> typecheckPat        ::  Doc -> TypeEnv -> Bool -> Pat -> Type -> TC (Type, TypeEnv)
> typecheckFormals    ::  Doc -> TypeEnv -> [Formal] -> TC ([Formal], TypeEnv)
> typecheckBinOp      ::  TypeEnv -> Type -> Type -> Type -> Expr -> Expr -> TC Type

> typecheck te e0 =
>   do
>     t0 <- typecheck' te e0
>     applytcsubst t0

> typecheck' te e0 =
>   case e0 of
>     Bool _ _     ->  return tBool
>     If e1 e2 e3  ->  do
>                        t1  <-  typecheck te e1
>                        unify (eloc e1) (text "in the condition of an if statement") te t1 (:<=:) tBool
>                        t2  <-  typecheck te e2
>                        t3  <-  typecheck te e3
>                        t4  <-  unify (eloc e2) (text "in the branches of an if statement") te t3 (:\/:) t2
>                        return t4
>     -- UnOp
>     BinOp e1 op e2
>                  ->  case op of
>                        Cat  ->  typecheckBinOp te tString tString tString e1 e2
>                        Add  ->  fresh >>= \ n -> typecheckBinOp te (TNatString n) (TNatString n) (TNatString n) e1 e2
>                        Sub  ->  typecheckBinOp te tNat tNat tNat e1 e2
>                        Mul  ->  typecheckBinOp te tNat tNat tNat e1 e2
>                        Div  ->  typecheckBinOp te tNat tNat tNat e1 e2
>                        Mod  ->  typecheckBinOp te tNat tNat tNat e1 e2
>                        And  ->  typecheckBinOp te tBool tBool tBool e1 e2
>                        Or   ->  typecheckBinOp te tBool tBool tBool e1 e2
>                        Leq  ->  fresh >>= \ n -> typecheckBinOp te (TNatString n) (TNatString n) tBool e1 e2
>                        Lt   ->  fresh >>= \ n -> typecheckBinOp te (TNatString n) (TNatString n) tBool e1 e2
>                        Geq  ->  fresh >>= \ n -> typecheckBinOp te (TNatString n) (TNatString n) tBool e1 e2
>                        Gt   ->  fresh >>= \ n -> typecheckBinOp te (TNatString n) (TNatString n) tBool e1 e2
>                        Neq  ->  fresh >>= \ n -> typecheckBinOp te (TNatString n) (TNatString n) tBool e1 e2
>                        Eq   ->  fresh >>= \ n -> typecheckBinOp te (TNatString n) (TNatString n) tBool e1 e2
>                        _    ->  tError unknown (text "unknown operator: " <> pretty op) tVoid
> 
>     Nat _ _      ->  return tNat
>     String _ _   ->  return tString
>
>     Var x        ->  case vLookup x te of
>                        Nothing  ->  tError (loc x) (text "unknown variable: " <> ltext x) tVoid
>                        Just t   ->  return t
>     Let ds e     ->  do
>                        -- non-recursive let
>                        ted <- typecheckDecls te ds
>                        typecheck (ted ++ te) e
>     Fun fs e     ->  do
>                        (fs',te')  <-  typecheckFormals (text "the declaration of an anonymous function") te fs
>                        t          <-  typecheck (te' ++ te) e -- |te'| does not contain the original environment 
>                        return (splitFormals fs' t)
>
>     RecFun x fs t e
>                  ->  do
>                        (fs',te')  <-  typecheckFormals (text "the declaration of a recursive function") te fs
>                        tr         <-  normalize te t >>= \ t' -> return (splitFormals fs' t')
>                        let b = SpVal x tr
>                        let tel = te' ++ (b : te)
>                        t'         <-  typecheck tel e
>                        unify (eloc e) (text "in the declaration of a recursive function") tel t' (:<=:) t
>                        applytcsubst tr
>     App e []     ->  typecheck te e
>     App e@(Con _) [ActExpr e0]
>                  ->  do
>                        t   <-  typecheckCon te e
>                        typecheckApp (eloc e) t e0
>     App e@(QualCon _ _) [ActExpr e0]
>                  ->  do
>                        t   <-  typecheckCon te e
>                        typecheckApp (eloc e) t e0
>     App e [ActExpr e0]
>                  ->  do
>                        t   <-  typecheck te e
>                        t'  <-  instantiate te t
>                        typecheckApp (eloc e) t' e0
>     App e [ActType ts]
>                  ->  do
>                        t   <-  typecheck te e
>                        reduceForall te t ts
>     App e (e0:es)
>                  ->  typecheck te (App (App e [e0]) es)
>     Rec p e      ->  do
>                        (t1,te1) <- typecheckPat (text "a recursive value") te False p tVoid
>                        t2 <- typecheck (te1 ++ te) e
>                        t1' <- unify (eloc e) (text "in the declaration of a recursive value") te t2 (:<=:) t1
>                        return t1'
>     Con x        ->  do
>                        t <- typecheckCon te e0
>                        case t of
>                          TyFun _ _  ->  tError (loc x) (text "constructors must not be partially applied") t
>                          _          ->  return t
>     Sel e l      ->  do
>                        t  <-  typecheck te e
>                        n  <-  fresh
>                        r  <-  fresh
>                        unify (eloc e) (text "in a record selection") te t (:<=:) (TyRecord [TyField (Just l) Nothing (TVar n)] (Just r))
>                        s  <-  gets tcsubst
>                        case lookup n s of
>                          Nothing  ->  {- tError (text "unification error in record selection") -} return (TVar n)
>                          Just t'  ->  return t'
>     Record rs    ->  do
>                        ts <- mapM (typecheckRow te) rs
>                        return (tyRecord ts)
>     Seq e1 e2    ->  do
>                        typecheck te e1
>                        typecheck te e2
>     Ref e        ->  do
>                        t <- typecheck te e
>                        return (TyApp tRef [ActType [t]])
>     Deref e      ->  do
>                        t  <-  typecheck te e
>                        n  <-  fresh
>                        let rt = (TyApp tRef [ActType [TVar n]])
>                        unify (eloc e) (text "in the dereference of a variable") te t (:<=:) rt
>                        s  <-  gets tcsubst
>                        case lookup n s of
>                          Nothing  ->  {- tError (text "unification error in dereference") -} return (TVar n)
>                          Just t'  ->  return t'
>     Assign e1 e2
>                  ->  do
>                        t1  <-  typecheck te e1
>                        t2  <-  typecheck te e2
>                        unify (eloc e1) (text "in an assignment") te t1 (:>=:) (TyApp tRef [ActType [t2]])
>                        return tUnit
>     While e1 e2  ->  do
>                        t1 <- typecheck te e1
>                        unify (eloc e1) (text "in the condition of a while statement") te t1 (:<=:) tBool
>                        t2 <- typecheck te e2
>                        unify (eloc e2) (text "in the body of a while statement") te t2 (:<=:) tUnit
>                        return tUnit
>     Case e arms  ->  do
>                        t <- typecheck te e
>                        foldM (typecheckArm (text "a case statement") te t) tVoid arms
>     Throw e      ->  do
>                        t  <-  typecheck te e
>                        unify (eloc e) (text "in a throw statement") te t (:<=:) tException
>                        return tVoid
>     Try e arms   ->  do
>                        t <- typecheck te e
>                        foldM (typecheckArm (text "a try statement") te tException) t arms
>     Continue e1 e2
>                  ->  do
>                        t1  <-  typecheck te e1
>                        n   <-  fresh
>                        let ct = (TyApp tCont [ActType [TVar n]])
>                        unify (eloc e1) (text "in a continue statement") te t1 (:<=:) ct
>                        t2  <-  typecheck te e2
>                        unify (eloc e2) (text "in the argument of a continuation") te t2 (:<=:) (TVar n)
>                        return tVoid
>     LetCC x t e  ->  do
>                        n   <-  fresh
>                        let ct = (TyApp tCont [ActType [TVar n]])
>                        unify (tloc t) (text "in the type of a letcc statement") te t (:<=:) ct
>                        rt  <-  typecheck (SpVal x ct : te) e
>                        unify (eloc e) (text "in the body of a letcc statement") te rt (:<=:) (TVar n)
>                        return rt
>     QualVar s x  ->  do
>                        te' <- normalizeStrExpr te s
>                        typecheck te' (Var x)
>     QualCon s x  ->  do
>                        t <- typecheckCon te e0
>                        case t of
>                          TyFun _ _  ->  tError (loc x) (text "constructors must not be partially applied") t
>                          _          ->  return t
>     -- Class
>     -- New
>     ArrFun e1 x e2
>                  ->  do
>                        t1  <-  typecheck te e1
>                        unify (eloc e1) (text "in an array construction") te t1 (:<=:) tNat
>                        t2  <-  typecheck (SpVal x tNat : te) e2
>                        return (TyApp tArray [ActType [t2]])
>     ArrList es   ->  do
>                        t <- foldM (\t0 e -> typecheck te e >>= \t -> unify (eloc e) (text "in an array construction") te t0 (:\/:) t >>= return) tVoid es
>                        return (TyApp tArray [ActType [t]])
>     Subscript e1 e2
>                  ->  do
>                        t1  <-  typecheck te e1
>                        n   <-  fresh
>                        let at = (TyApp tArray [ActType [TVar n]])
>                        unify (eloc e1) (text "in the lhs of an array subscription") te t1 (:<=:) at
>                        t2  <-  typecheck te e2
>                        unify (eloc e2) (text "in the rhs of an array subscription") te t2 (:<=:) tNat
>                        s   <-  gets tcsubst
>                        case lookup n s of
>                          Nothing  ->  tError (eloc e2) (text "unification error in array subscription") tVoid
>                          Just t'  ->  return t'
>     Typed e t0   ->  do
>                        t1 <- typecheck te e
>                        t0' <- normalize te t0
>                        unify (eloc e) (text "in an explicitly typed expression") te t1 (:<=:) t0'
>                        return t0'
>     _            ->  tError (eloc e0) (text "unimplemented construct") tVoid
>
>   where
>     typecheckCon te (Con x) =
>                      case vLookup x te of
>                        Nothing  ->  tError (loc x) (text "unknown constructor: " <> pretty x) tVoid
>                        Just t   ->  instantiate te t
>     typecheckCon te (QualCon s x) =
>                      do
>                        te' <- normalizeStrExpr te s
>                        typecheckCon te' (Con x)
>     typecheckApp l t e0 =
>                      do
>                        t0  <-  typecheck te e0
>                        n   <-  fresh
>                        let ft = tFun [t0] (TVar n)
>                        unify l (text "in a function application") te t (:<=:) ft
>                        s   <-  gets tcsubst
>                        case lookup n s of
>                          Nothing   ->  -- tError (text "unification error in application") tVoid
>                                        return tVoid -- can happen if the function returns Void
>                          Just t'   ->  return t'


Helper function to split a list of |Formal|s:

> splitFormals :: [Formal] -> Type -> Type
> splitFormals fs  =  foldr (\g f -> transformGroup g . f) id (groupBy groupFormal fs)
>   where
>     groupFormal x y = formalClass x == formalClass y
>
>     formalClass (ForPat _)      =  False
>     formalClass (ForTypePat _)  =  True
>
>     transformGroup g@((ForPat _) : _)      =  TyFun g
>     transformGroup g@((ForTypePat _) : _)  =  TyForall g

Helper function to turn formal arguments into actual arguments:

> formals2actuals :: [Formal] -> [Actual]
> formals2actuals = map formal2actual
>   where
>     formal2actual (ForPat (PatVar x))  =  ActExpr (Var x)
>     formal2actual (ForTypePat tvs)     =  ActType (map TyVar tvs)

> typecheckDecl te0 d =
>   case d of
>     Val p e      ->  do
>                        t         <-  typecheck te0 e
>                        (_, te1)  <-  typecheckPat (text "a value declaration") te0 False p t
>                        return te1
>     Function rec
>                  ->  do
>                        tformals  <-  mapM (typecheckFormals (text "a function declaration") te0) (map (\ (x, fs, t, e) -> fs) rec)
>                        te1       <-  zipWithM (\ (x, _, t, e) (fs, _) -> normalize te0 t >>= \ t' -> return (SpVal x (splitFormals fs t'))) rec tformals
>                        -- tError (pretty te1) ()
>                        let te2 = te1 ++ te0
>                        zipWithM_  (\ (x, fs, t, e) (_, te') -> 
>                                      do
>                                        -- tError (text "final: " <> pretty (te' ++ te2)) ()
>                                        t'  <-  typecheck (te' ++ te2) e
>                                        unify (eloc e) (text "in a function declaration") (te' ++ te2) t' (:<=:) t ) rec tformals
>                        te3       <-  mapM (\ (SpVal x t) -> applytcsubst t >>= return . SpVal x) te1
>                        return te3
>     Type x fs t  ->  do
>                        -- TODO: handle type declarations with formal parameters
>                        t'   <-  normalize te0 t
>                        -- n <- fresh
>                        let y     =  x --  makeTyName n x
>                        t''  <-  normalize te0 (TyFun fs t')
>                        return [SpType y t'']
>     Datatype x fs alts
>                  ->  do
>                        -- TODO: add formals to type environment
>                        let acts  =   formals2actuals fs
>                        n <- fresh
>                        let y     =   (unknown,makeTyName n (snd x))
>                        ts        <-  mapM (\a -> typecheckAlt (SpData y : te0) a (TyApp (TyCon y) acts)) alts
>                        let cons  =   map (\ (SpCon c t) -> SpCon c (splitFormals fs t)) ts
>                        return (reverse (SpData y : cons))  -- order doesn't matter, but this will print it nicely in interactive mode
>     Data x       ->  return [SpData x] -- assign unique number, or is it assigned already?
>     Constr c t   ->  do
>                        t'        <-  normalize te0 t
>                        return [SpCon c t']
>     Exception (c,Nothing)
>                  ->  return [SpException (c,Nothing)]
>     Exception (c,Just t)
>                  ->  do
>                        t'        <-  normalize te0 t
>                        return [SpException (c, Just t')]
>     Module m [] s
>                  ->  do
>                        te        <-  normalizeStrExpr te0 s
>                        return [SpModule m (Sig te)]
>     Open s       ->  normalizeStrExpr te0 s
>     Local ds1 ds2
>                  ->  do
>                        te1 <- typecheckDecls te0 ds1
>                        typecheckDecls (te1 ++ te0) ds2
>     -- Interface
>     _            ->  tError unknown (text "unimplemented declaration construct") []

> typecheckDecls te0 ds =
>   foldM ( \ te d -> typecheckDecl (te ++ te0) d >>= \ te' -> return (te' ++ te)) [] ds

> typecheckRow te0 (Field l e) =
>   do
>     t <- typecheck te0 e
>     return (TyField l Nothing t)

> typecheckArm sit te0 ht rt (p,e) =
>   do
>     -- tError (text "context type: " <> ppType ht) ()
>     (_, te)  <-  typecheckPat (text "an arm of" <+> sit) te0 True p ht
>     ht'      <-  applytcsubst ht
>     -- tError (text "applied context type: " <> ppType ht') ()
>     -- tError (text "checking result: " <> ppType tt) ()
>     t        <-  typecheck (te ++ te0) e
>     t'       <-  unify (eloc e) (text "in the arm of" <+> sit) te0 rt (:\/:) t
>     return t'

> typecheckAlt te0 (c,Nothing)  tr  =  return (SpCon c tr)
> typecheckAlt te0 (c,Just t)   tr  =
>   normalize te0 t >>= \t' -> return (SpCon c (tFun [t'] tr))

The |multi| argument indicates whether we're checking multiple patterns (a case)
or just a single one (a fun or function declaration). In the former case we introduce
a row variable for records, to allow record inference through case patterns. If we'd
do this also for fun and function, we'd get spurious row variables showing up in the
types of functions ...

> typecheckPat sit te multi p tc = normalize te tc >>= \ tc' -> typecheckPat' sit te multi p tc'

> typecheckPat' sit te multi p tc =
>   case p of
>     PatSucceed       ->  do
>                            n   <-  fresh
>                            t'  <-  unify unknown (text "in a succeed pattern" <+> sit) te tc (:<=:) (TVar n)
>                            return (t', []) -- was: return tc, but that'd return Void?
>     PatVar x         ->  do
>                            n   <-  fresh
>                            t'  <-  unify (loc x) (text "in a variable pattern" <+> sit) te tc (:<=:) (TVar n) 
>                            return (t', [SpVal x t'])
>                          -- return (tc, [SpVal x tc])
>     PatCon x p       ->  case vLookup x te of
>                            Nothing  ->  tError (loc x) (text "unknown constructor: " <> pretty x) (tc, [])
>                            Just tp  ->
>                                       do
>                                         t <- instantiate te tp -- constructors are not truly polymorphic
>                                         case p of
>                                           Nothing  ->  do
>                                                          t'  <-  unify (loc x) (text "in a constructor pattern" <+> sit) te tc (:<=:) t
>                                                          return (t', [])
>                                           Just p'  ->  do
>                                                          m   <-  fresh  -- argument type
>                                                          n   <-  fresh  -- result type
>                                                          unify (loc x) (text "in a constructor pattern" <+> sit) te t (:<=:) (tFun [TVar m] (TVar n))
>                                                          s   <-  gets tcsubst
>                                                          pt  <-  applytcsubst (TVar m)
>                                                          rt  <-  applytcsubst (TVar n)
>                                                          (_, te')  <-  typecheckPat sit te multi p' pt
>                                                          unify (loc x) (text "in a constructor pattern" <+> sit) te tc (:<=:) rt 
>                                                          return (rt, te')
>     PatTyped p0 t0   ->  do
>                            unify (tloc t0) (text "in an explicitly typed pattern" <+> sit) te tc (:<=:) t0 -- TODO: think about this
>                            (t1,te1)  <-  typecheckPat sit te multi p0 t0
>                            return (t1,te1)
>     PatRecord rs     ->  do
>                            rs'  <-  mapM (typecheckPatRow sit te multi) rs
>                            let (trs, tes) = unzip rs'
>                            rv   <-  if multi then fmap Just fresh else return Nothing
>                            let r = TyRecord trs rv
>                            r'   <-  unify (locPatRows rs) (text "in a record pattern" <+> sit) te tc (:<=:) r
>                            -- tError (text "returning " <> ppType r' <> text " from unifying " <> ppType r <> text " with context " <> ppType tc) ()
>                            return (r', concat tes)
>     PatAlt p1 p2     ->  do
>                            (t1,te1)  <-  typecheckPat sit te multi p1 tc
>                            (t2,te2)  <-  typecheckPat sit te multi p2 tc
>                            t3 <- unify (locPat p1) (text "in an or-pattern" <+> sit) te t1 (:/\:) t2
>                            te3 <- unifyTypeEnvs (locPat p1) sit te te1 te2
>                            return (t2, te3)
>     PatBoth p1 p2    ->  do
>                            (t1,te1)  <-  typecheckPat sit te multi p1 tc
>                            (t2,te2)  <-  typecheckPat sit te multi p2 tc
>                            t3 <- unify (locPat p1) (text "in an and-pattern" <+> sit) te t1 (:/\:) t2
>                            return (t3, te1 ++ te2)
>     _                ->  do
>                            tError unknown (text "unimplemented pattern construct") ()
>                            return (tc, [])

> typecheckPatRow sit te0 multi (Field l p)  =
>   do
>     (t, te) <- typecheckPat sit te0 multi p tVoid
>     return (TyField l Nothing t, te)

> typecheckFormals sit te []      =  return ([], [])
> typecheckFormals sit te (f:fs)  =
>   case f of
>     ForPat p        ->  do
>                           n          <-  fresh
>                           (t1,te1)   <-  typecheckPat sit te False p (TVar n)
>                           (fs2,te2)  <-  typecheckFormals sit (te1 ++ te) fs
>                           return (ForPat (PatTyped (PatVar (unknown,"_")) t1) : fs2, te1 ++ te2) -- we could leave the variable name in here to allow dependent types
>     ForTypePat tvs  ->  do
>                           (fs1,te1)  <-  typecheckFormals sit (map SpData tvs ++ te) fs
>                           return (ForTypePat tvs : fs1, te1)

> typecheckBinOp te t1 t2 t0 e1 e2 =
>   do
>     t1' <- typecheck te e1
>     t2' <- typecheck te e2
>     unify (eloc e1) (text "in the lhs of an operator application") te t1' (:<=:) t1
>     unify (eloc e2) (text "in the rhs of an operator application") te t2' (:<=:) t2
>     return t0

Converting the TC monad into IO.

> runTC :: Options -> Int -> TC a -> IO (Maybe (a, Int))
> runTC o cnt x
>   | typechecker o  =  
>     case runState x (TCState cnt [] []) of
>       (y,s) -> mapM_ (putStrLn . render (Page (pagewidth o)) . pretty . Error) (reverse (tcerrors s)) >> return (Just (y, tccount s))
>   | otherwise      =  return Nothing

Unification is now parameterized by an |UOrdering|. The |UOrdering|
indicates what we're interested in: |:<=:| and |:>=:| are for normal
subsumption checks, |:\/:| is for actual unification (used in the
branches of an |if| and a |case|), and |:/\:| is anti-unification
(used to unify function arguments).

Polymorphism is treated rather ad-hoc (and could be buggy in advanced
cases), but record subtyping should respect the |UOrdering| properly
...

For |:\/:| and |:/\:|, we return the unified and anti-unified type, 
respectively.

> data UOrdering = (:<=:) | (:==:) | (:\/:) | (:/\:) | (:>=:) deriving (Eq)

> flipU :: UOrdering -> UOrdering
> flipU (:==:) = (:==:)
> flipU (:<=:) = (:>=:)
> flipU (:>=:) = (:<=:)
> flipU (:\/:) = (:/\:)
> flipU (:/\:) = (:\/:)

> unify :: Loc -> Doc -> TypeEnv -> Type -> UOrdering -> Type -> TC Type
> unify l sit te0 r1 (:>=:) r2 = unify l sit te0 r2 (:<=:) r1  -- hopefully unnecessary
> unify l sit te0 r1 dir r2 = do
>                    s1 <- normalize te0 r1
>                    s2 <- normalize te0 r2
>                    unify' te0 s1 s2
>   where
>     unify' :: TypeEnv -> Type -> Type -> TC Type
>     unify' te s1 s2 =
>       case (s1,s2) of
>         -- it's important that Value/Void cases are handled before variable bindings
>         (TyCon (_,c), t)
>            | c == iValue && dir `elem` [(:>=:),(:/\:)]
>                                ->  return t -- TODO: this seems somewhat fragile
>            | c == iVoid && dir `elem` [(:<=:),(:\/:)]
>                                ->  return t
>         (t, TyCon (_,c))
>            | c == iValue && dir `elem` [(:<=:),(:/\:)]
>                                ->  return t
>            | c == iVoid && dir `elem` [(:>=:),(:\/:)]
>                                ->  return t
>         (_, TVar n)
>            | not (n `elem` tvars s1)
>                                ->  bind n s1 >> return s1
>         (TVar n, _)
>            | not (n `elem` tvars s2)
>                                ->  bind n s2 >> return s2
>         (TVar m, TVar n)
>            | m == n            ->  return (TVar m)
>         (TNatString m, TNatString n)
>            | m == n            ->  return (TNatString m)
>            | otherwise         ->  bind m s2 >> return s2
>         (TyCon (_,c1), TNatString m)
>            | c1 == iString || c1 == iNat
>                                ->  bind m s1 >> return s1
>         (TNatString m, TyCon (_,c2))
>            | c2 == iString || c2 == iNat
>                                ->  bind m s2 >> return s2
>         (x@(TyVar (_,v1)), TyVar (_,v2))
>            | v1 == v2          ->  return x
>                                    -- TODO: we should check if the var is in scope
>         (x@(TyCon (_,c1)), TyCon (_,c2))
>            | c1 == c2          ->  return x
>                                    -- TODO: same as above
>         (TyApp t1 ts1, TyApp t2 ts2)
>                                ->  do
>                                      t3 <- unify' te t1 t2
>                                      t3' <- applytcsubst t3
>                                      t3'' <- normalize te t3'
>                                      let dir' = case t3'' of
>                                                   TyCon (_,c)
>                                                     | c == iRef        ->  (:==:)
>                                                     | c == iArray      ->  dir
>                                                     | c == iCont       ->  flipU dir
>                                                   _                    ->  (:==:) -- assume invariance
>                                      ts3 <- zipWithM (\x y -> unifyActual l sit te x dir' y) ts1 ts2
>                                      return (TyApp t3 ts3)
>         (TyFun fs1 t1, TyFun fs2 t2)
>                                ->  do -- the problem is that function arguments can be grouped in multiple |TyFun|s in incompatible ways
>                                      let  (fs1':fs1r)  =  fs1  -- both lists aren't empty, because the types are normalized
>                                           (fs2':fs2r)  =  fs2
>                                      (s,fs3) <- unifyFormal l sit te fs1' (flipU dir) fs2'  -- contravariance
>                                      t3 <- unify l sit te (TyFun fs1r t1) dir (subst s [] (TyFun fs2r t2))  -- |unify|, not |unify'|, to normalize again
>                                      return (TyFun [fs3] t3)
>         (TyForall fs1 t1, TyForall fs2 t2)
>                                ->  do
>                                      let  (fs1':fs1r)  =  fs1
>                                           (fs2':fs2r)  =  fs2
>                                      (s,fs3) <- unifyFormal l sit te fs1' (flipU dir) fs2'
>                                      t3 <- unify l sit te (TyForall fs1r t1) dir (subst s [] (TyForall fs2r t2))
>                                      return (TyForall [fs3] t3)
> {-
>         (TyForall fs1 t1, _)
>            | dir `elem` [(:<=:),(:\/:)]
>                                ->  do
>                                      s <- mapM (\f -> fresh >>= \n -> return (f, TVar n)) (concatMap (\ (ForTypePat fs) -> fs) fs1)
>                                      unify te (subst s [] t1) dir s2
>         (_, TyForall fs2 t2)
>            | dir `elem` [(:>=:),(:\/:)]
>                                ->  do
>                                      s <- mapM (\f -> fresh >>= \n -> return (f, TVar n)) (concatMap (\ (ForTypePat fs) -> fs) fs2)
>                                      unify te s1 dir (subst s [] t2)
> -}
>         (TyRecord rs1 rv1, TyRecord rs2 rv2)
>                                ->  unifyRows te rs1 rv1 rs2 rv2
>         _                      ->  let (t1,t2,x1,x2,e) | dir == (:\/:)  =  ("unify","and",s1,s2,tVoid)
>                                                        | dir == (:/\:)  =  ("anti-unify","and",s1,s2,tValue)
>                                                        | dir == (:<=:)  =  ("instantiate","to",s1,s2,tVoid)
>                                                        | dir == (:>=:)  =  ("instantiate","to",s2,s1,tVoid)
>                                                        | dir == (:==:)  =  ("equate","and",s1,s2,tVoid)
>                                    in tError l (group (text "cannot " <> text t1 <+> ppType x1 <> text " " <> text t2 <+> ppType x2 <+> sit)) e
> 
>     unifyRows te rs1 rv1 rs2 rv2 =
>         let
>           rs1'  =  sortRows rs1
>           rs2'  =  sortRows rs2
>         in
>           do
>             rs <- mergeRows rs1' rs2'
>             let finalRows  =  map fst rs
>             let isLeft (Just (:<=:))   =  True
>                 isLeft _               =  False
>                 isRight (Just (:>=:))  =  True
>                 isRight _              =  False
>                 leftRows   =  (map fst . filter (isLeft . snd)) rs
>                 rightRows  =  (map fst . filter (isRight . snd)) rs
>             m <- case dir of
>                    (:<=:)  ->  return rv2
>                    (:>=:)  ->  return rv1
>                    (:/\:)  ->  fmap Just fresh
>                    _       ->  return Nothing
>             case rv1 of
>               Just n1 | rv1 /= m  ->  bind n1 (TyRecord rightRows m)
>               _                   ->  return ()
>             case rv2 of
>               Just n2 | rv2 /= m  ->  bind n2 (TyRecord leftRows m)
>               _                   ->  return ()
>             return (TyRecord finalRows m)
>
>       where
>         dirs = dir : maybe [] (const [(:>=:)]) rv1 ++ maybe [] (const [(:<=:)]) rv2
>         mergeRows (r1@(TyField l1 _ t1) : rs1'') (r2@(TyField l2 _ t2) : rs2'')
>           | l1 == l2  =  do
>                            t3   <-  unify l sit te t1 dir t2
>                            rs3  <-  mergeRows rs1'' rs2''
>                            return ((TyField l1 Nothing t3, Nothing) : rs3)
>           | l1 <= l2  =  singleRow r1 (:<=:) rs1'' (r2:rs2'')
>           | l2 <= l1  =  singleRow r2 (:>=:) (r1:rs1'') rs2''
>         mergeRows (r:rs) []  =  singleRow r (:<=:) rs []
>         mergeRows [] (r:rs)  =  singleRow r (:>=:) rs []
>         mergeRows [] []      =  return []
>
>         singleRow (r@(TyField (Just l) _ _)) dir' rs1 rs2
>           | any (== (:\/:)) dirs              =  mergeRows rs1 rs2
>           | any (`elem` [(:/\:), dir']) dirs  =  fmap ((r, Just dir') :) (mergeRows rs1 rs2)
>           | otherwise                         =  do
>                                                    tError unknown (text "could not find label " <> pretty l) ()
>                                                    fmap ((r, Just dir') :) (mergeRows rs1 rs2)
>

> unifyActual :: Loc -> Doc -> TypeEnv -> Actual -> UOrdering -> Actual -> TC Actual
> unifyActual l sit te a1 dir a2 =
>       case (a1,a2) of
>         (ActType ts1, ActType ts2)
>                                ->  fmap ActType $ zipWithM (\x y -> unify l sit te x dir y) ts1 ts2
>         _                      ->  tError l (group (text "cannot unify value-level arguments (forgot to write <> around a type argument?)" <+> sit)) (ActType [])
>

> unifyFormal :: Loc -> Doc -> TypeEnv -> Formal -> UOrdering -> Formal -> TC (TypeSubst,Formal)
> unifyFormal l sit te f1 dir f2 =
>       case (f1,f2) of
>         (ForPat (PatTyped _ t1), ForPat (PatTyped _ t2))
>                                ->  do
>                                      t3 <- unify l sit te t1 dir t2
>                                      return ([], ForPat (PatTyped (PatVar (unknown,"_")) t3))
>         (ForTypePat ts1, ForTypePat ts2)
>            | length ts1 == length ts2
>                                ->  do
>                                      let s = zipWith (\x y -> (x, TyVar y)) ts2 ts1
>                                      return (s, ForTypePat ts1)
>         _                      ->  tError l (group (text "cannot unify" <+> ppFormal f1 <> text " and" <+> ppFormal f2 <+> sit)) ([], ForTypePat [])


> sortRows :: [TyRow] -> [TyRow]
> sortRows = sortBy (\ (TyField x _ _) (TyField y _ _) -> compare x y) . numberRows' 1
>   where
>     numberRows' n []                          =  []
>     numberRows' n (TyField Nothing v t : rs)  =  TyField (Just (NumLab n)) v t : numberRows' (n+1) rs
>     numberRows' n (r : rs)                    =  r : numberRows' (n+1) rs

> findRow :: Lab -> [TyRow] -> Maybe Type
> findRow _ []        =  Nothing
> findRow (AlNumLab x) (TyField (Just (AlNumLab y)) _ t : rs)
>   | x == y          =  Just t
> findRow (NumLab m) (TyField (Just (NumLab n)) _ t : rs)
>   | m == n          =  Just t
> findRow (NumLab 1) (TyField Nothing _ t : rs)
>                     =  Just t
> findRow (NumLab m) (TyField Nothing _ t : rs)
>                     =  findRow (NumLab (m-1)) rs
> findRow l (_ : rs)  =  findRow l rs

> unifyTypeEnvs :: Loc -> Doc -> TypeEnv -> TypeEnv -> TypeEnv -> TC TypeEnv
> unifyTypeEnvs l sit te0 te1 te2 = unifyTypeEnvs' (sortTypeEnv te1) (sortTypeEnv te2)
>   where
>     sortTypeEnv = sortBy compareTypeEnv
>     compareTypeEnv (SpVal (_,n1) _) (SpVal (_,n2) _) = compare n1 n2
>     compareTypeEnv _ _ = EQ
>     unifyTypeEnvs' (SpVal x@(_,n1) t1 : te1') (SpVal (_,n2) t2 : te2')
>       | n1 == n2 = do  s   <-  unify l (text "in an or-pattern" <+> sit) te0 t1 (:\/:) t2 
>                        rs  <-  unifyTypeEnvs' te1' te2'
>                        return (SpVal x s:rs)
>     unifyTypeEnvs' [] [] = return []
>     unifyTypeEnvs' _ _ = tError l (text "incompatible or-patterns" <+> sit) []

Instantiates a (top-level) quantified type with fresh type variables.

> instantiate :: TypeEnv -> Type -> TC Type
> instantiate te0 t0 =
>   do
>     t1 <- normalize te0 t0
>     case t1 of
>       TyForall fs2 t2 ->
>         do
>           s <- mapM  (\f -> fresh >>= \n -> return (f, TVar n))
>                      (concatMap (\ (ForTypePat fs) -> fs) fs2)
>           let t3 = subst s [] t2
>           instantiate te0 t3
>       _ -> return t1

This function determines the unification variables in a type.

> tvars :: Type -> [Int]
> tvars t =
>   case t of
>     TyApp t ts     ->  tvars t ++ concatMap tvarsActual ts
>     TyFun ts t     ->  concatMap tvarsFormal ts ++ tvars t
>     TyForall ts t  ->  concatMap tvarsFormal ts ++ tvars t
>     TyRecord tr Nothing
>                    ->  concatMap tvarsRow tr
>     TyRecord tr (Just n)
>                    ->  concatMap tvarsRow tr ++ [n]
>     TyClass ts     ->  concatMap tvarsSlot ts
>     TyObject ts    ->  concatMap tvarsSlot ts
>     TVar n         ->  [n]
>     TNatString n   ->  [n]
>     _              ->  []
>   where
>     tvarsActual :: Actual -> [Int]
>     tvarsActual (ActType ts)  =  concatMap tvars ts
>     tvarsActual _             =  []
>
>     tvarsFormal :: Formal -> [Int]
>     tvarsFormal (ForPat p)       =  tvarsPat p
>     tvarsFormal _                =  []
>
>     tvarsRow (TyField _ _ t)  =  tvars t
>
>     tvarsSlot (TyMethod _ t)  =  tvars t
>
>     tvarsPat (PatTyped _ t)  =  tvars t
>     tvarsPat _               =  []

This function determines the location of a type.

> tloc :: Type -> Loc
> tloc t =
>   case t of
>     TyApp t _      ->  tloc t
>     TyFun [] t     ->  tloc t
>     TyFun (t:_) _  ->  locFormal t
>     TyForall [] t  ->  tloc t
>     TyForall (t:_) _
>                    ->  locFormal t
>     TyRecord tr _  ->  locTyRows tr
>     TyClass []     ->  unknown
>     TyClass (s:_)  ->  locSlot s
>     TyObject []    ->  unknown
>     TyObject (s:_) ->  locSlot s
>     _              ->  unknown

> 
> locFormal (ForPat p)      =  locPat p
> locFormal (ForTypePat []) =  unknown
> locFormal (ForTypePat ((l,_):_)) = l

> locTyRows []                 =  unknown
> locTyRows (TyField _ _ t:_)  =  tloc t
>
> locPatRows []               =  unknown
> locPatRows (Field _ p:_)    =  locPat p

> locExpRows []               =  unknown
> locExpRows (Field _ e:_)    =  eloc e
>
> locSlot _                 =  unknown -- preliminary

> locPat PatSucceed         =  unknown
> locPat (PatVar (l,_))     =  l
> locPat (PatCon (l,_) _)   =  l
> locPat (PatTyped _ t)     =  tloc t
> locPat (PatAlt p _)       =  locPat p
> locPat (PatBoth p _)      =  locPat p
> locPat _                  =  unknown

> eloc :: Expr -> Loc
> eloc e =
>   case e of
>     Bool l _ -> l
>     If _ e _ -> eloc e
>     UnOp _ e -> eloc e
>     BinOp e _ _ -> eloc e
>     Nat l _ -> l
>     String l _ -> l
>     Var (l,_) -> l
>     Let _ e -> eloc e
>     Fun [] e -> eloc e
>     Fun (f:_) _ -> locFormal f
>     RecFun (l,_) _ _ _ -> l
>     App e _ -> eloc e
>     Rec p _ -> locPat p
>     Con (l,_) -> l
>     Sel e _ -> eloc e
>     Record r -> locExpRows r
>     Seq e1 e2 -> eloc e2
>     Ref e -> eloc e
>     Deref e -> eloc e
>     Assign e _ -> eloc e
>     While e _ -> eloc e
>     Case e _ -> eloc e
>     Throw e -> eloc e
>     Try e _ -> eloc e
>     Continue e _ -> eloc e
>     LetCC (l,_) _ _ -> l
>     QualVar _ (l,_) -> l
>     QualCon _ (l,_) -> l
>     Class [] -> unknown
>     Class (s:_) -> locSlot s
>     New e -> eloc e
>     ArrFun e _ _ -> eloc e
>     ArrList [] -> unknown
>     ArrList (e:_) -> eloc e
>     Subscript e _  -> eloc e
>     Typed _ t -> tloc t
>     _ -> unknown

The function |normalize| is a sanity check for types.  It performs
lookups in modules. It could also be used to perform kind
checking. It's a bit fuzzy what it really does, which should be
clarified.

> normalize :: TypeEnv -> Type -> TC Type
> normalize te t0 =
>   case t0 of
>     QualTyCon s x  ->  do
>                          te' <- normalizeStrExpr te s
>                          normalize te' (TyCon x)
>     TyCon x        ->  case tLookup x te of
>                          Nothing  ->  tError (loc x) (text "unknown type: " <> pretty x) tVoid
>                          Just t   ->  return t
>     TyApp t []     ->  normalize te t
>     TyApp t xs     ->  do
>                          t'   <-  normalize te t
>                          xs'  <-  mapM (normalizeActual te) xs
>                          case (t',xs') of
>                            (TyFun (ForTypePat _ : _) _, ActType t0 : ys) ->
>                              do
>                                t'' <- reduceFun te t' t0
>                                normalize te (TyApp t'' ys)
>                            _ -> return (TyApp t' xs')
>     TyFun [] t     ->  normalize te t
>     TyForall [] t  ->  normalize te t
>     TyRecord rs rv
>                    ->  do
>                          s <- gets tcsubst
>                          let cont = fmap (\r -> TyRecord r rv) $ mapM (normalizeRow te) rs
>                          case rv of
>                            Just n ->
>                              case lookup n s of
>                                Just (TyRecord rs' rv')  ->  normalize te (TyRecord (rs ++ rs') rv')
>                                Nothing                  ->  cont
>                            Nothing -> cont
>     TVar n         ->  do
>                          s <- gets tcsubst
>                          case lookup n s of
>                            Just t'  ->  normalize te t'
>                            _        ->  return t0
>     TNatString n   ->  do
>                          s <- gets tcsubst
>                          case lookup n s of
>                            Just t'  ->  normalize te t'
>                            _        ->  return t0
>     _              ->  return t0

> normalizeActual :: TypeEnv -> Actual -> TC Actual
> normalizeActual te (ActType ts)  =  fmap ActType (mapM (normalize te) ts)
> normalizeActual te t             =  return t

> normalizeRow :: TypeEnv -> TyRow -> TC TyRow
> normalizeRow te (TyField l v t) = normalize te t >>= \ t' -> return (TyField l v t')

> normalizeStrExpr :: TypeEnv -> StrExpr -> TC TypeEnv
> normalizeStrExpr te s =
>   case s of
>     StrVar v       ->  case strLookup v te of
>                          Nothing   ->  tError (loc v) (text "unknown struct: " <> pretty v) te
>                          Just te'  ->  return te'
>     StrSel t v     ->  do
>                          te' <- normalizeStrExpr te t
>                          normalizeStrExpr te' (StrVar v)
>     Struct ds      ->  do
>                          te' <- typecheckDecls te ds
>                          return te'
>     StrApp s []    ->  normalizeStrExpr te s
>     -- StrFunctor
>     -- StrApp    

Reduction on types.

> reduceForall :: TypeEnv -> Type -> [Type] -> TC Type
> reduceForall te t0 []  =  normalize te t0
> reduceForall te t0 ts  =
>     do
>       t1 <- normalize te t0
>       case t1 of
>         TyForall (ForTypePat fs0:fsr) t2  -> reduce' reduceForall TyForall te fs0 fsr t2 ts
>         _                                 ->
>           tError unknown (group (text "illegal application of" <+> ppType t0 <> text " to type arguments")) t0

> reduceFun :: TypeEnv -> Type -> [Type] -> TC Type
> reduceFun te t0 []  =  normalize te t0
> reduceFun te t0 ts  =
>     do
>       t1 <- normalize te t0
>       case t1 of
>         TyFun (ForTypePat fs0:fsr) t2     -> reduce' reduceFun TyFun te fs0 fsr t2 ts
>         _                                 ->
>           tError unknown (group (text "illegal application of" <+> ppType t0 <> text " to type arguments")) t0


> reduce' rec con te fs0 fsr t2 ts =
>         do
>           let  m           =  min (length fs0) (length ts)
>                (ts1,tsr1)  =  splitAt m ts
>                (fs1,fsr1)  =  splitAt m fs0
>           ts1' <- mapM (normalize te) ts1
>           let  t           =  subst (zip fs1 ts1') [] t2
>           rec te (con ((if null fsr1 then id else (ForTypePat fsr1 :)) fsr) t) tsr1


Substitution on types.

> subst :: TypeSubst -> TSubst -> Type -> Type
> subst s1 s2 t0 = -- Debug.Trace.trace (render (Page 80) (pretty t0)) $
>   case t0 of
>     TVar n         ->  case lookup n s2 of
>                          Nothing  ->  t0
>                          Just t'  ->  subst s1 s2 t'
>     TNatString n   ->  case lookup n s2 of
>                          Nothing  ->  t0
>                          Just t'  ->  subst s1 s2 t'
>     TyVar v        ->  case tcLookup v s1 of
>                          Nothing  ->  t0
>                          Just t'  ->  t'
>     TyApp t1 tas   ->  TyApp (subst s1 s2 t1) (substActuals s1 s2 tas)
>     TyFun fs t     ->  let  (fs',s') = substFormals s1 s2 fs
>                        in   TyFun fs' (subst s' s2 t)
>     TyForall fs t  ->  let  (fs',s') = substFormals s1 s2 fs
>                        in   TyForall fs' (subst s' s2 t)
>     TyRecord trs Nothing
>                    ->  TyRecord (substRows s1 s2 trs) Nothing
>     TyRecord trs (Just n)
>                    ->  case lookup n s2 of
>                          Nothing                  ->  TyRecord (substRows s1 s2 trs) (Just n)
>                          Just (TyRecord trs' n')  ->  subst s1 s2 (TyRecord (trs ++ trs') n')
>     -- TyPred
>     TyClass ts     ->  TyClass (substSlots s1 s2 ts)
>     TyObject ts    ->  TyObject (substSlots s1 s2 ts)
>     _              ->  t0
>   where
>     substActuals :: TypeSubst -> TSubst -> [Actual] -> [Actual]
>     substActuals s1 s2 = map subActual
>       where
>         subActual (ActType ts)  =  ActType (map (subst s1 s2) ts)
>         subActual t             =  t
>
>     substFormals :: TypeSubst -> TSubst -> [Formal] -> ([Formal], TypeSubst)
>     substFormals s1 s2 = foldl (\ (fs,s) f ->  let  (f',s') = substFormal s s2 f
>                                                in   (fs ++ [f'], s')) ([],s1)
>                        
>       where
>         substFormal :: TypeSubst -> TSubst -> Formal -> (Formal, TypeSubst)
>         substFormal s1 s2 (ForPat (PatTyped x t))  =  (  ForPat (PatTyped x (subst s1 s2 t)),s1)
>         substFormal s1 s2 (ForTypePat vs)          =  (  ForTypePat vs,
>                                                          filter (\ ((_,v),_) -> not (elem v (map snd vs))) s1)
>         substFormal s1 s2 f                        =  (  f, s1  )
>
>     substRows :: TypeSubst -> TSubst -> [TyRow] -> [TyRow]
>     substRows s1 s2 = map substRow
>       where
>         substRow (TyField l v t) = TyField l v (subst s1 s2 t)
>
>     substSlots :: TypeSubst -> TSubst -> [TySlot] -> [TySlot]
>     substSlots s1 s2 = map substSlot
>       where
>         substSlot (TyMethod v t) = TyMethod v (subst s1 s2 t)


> stdtypeenv :: TypeEnv
> stdtypeenv
>   =  [  SpData (unknown,iBool),
>         SpData (unknown,iNat),
>         SpData (unknown,iString),
>         SpData (unknown,iException),
>         SpData (unknown,iRef),
>         SpData (unknown,iArray),
>         SpData (unknown,iCont),
>         SpData (unknown,iVoid),
>         SpData (unknown,iValue),
>         SpException ((unknown,"Div"), Nothing),
>         SpException ((unknown,"Match"), Nothing),
>         SpException ((unknown,"Mod"), Nothing),
>         SpException ((unknown,"Sub"), Nothing),
>         SpVal (unknown,"true") tBool,
>         SpVal (unknown,"false") tBool,
>         SpVal (unknown,"write") (tFun [tString] tUnit),
>         SpVal (unknown,"show") (tFun [tValue] tString),
>         SpVal (unknown,"size") (tFun [TyApp tArray [ActType [tValue]]] tNat),
>         SpVal (unknown,"putchar") (tFun [tString] tUnit),
>         SpVal (unknown,"getchar") (tFun [tUnit] tString),
>         SpVal (unknown,"chr") (tFun [tNat] tString),
>         SpVal (unknown,"ord") (tFun [tString] tNat)  ]

